package io.annaclemens.xivchat.model

import android.annotation.SuppressLint
import android.content.Context
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.ImageView
import android.widget.ProgressBar
import android.widget.TextView
import androidx.lifecycle.LiveData
import androidx.lifecycle.Observer
import androidx.recyclerview.widget.RecyclerView
import androidx.room.Dao
import androidx.room.Delete
import androidx.room.Entity
import androidx.room.Insert
import androidx.room.OnConflictStrategy
import androidx.room.PrimaryKey
import androidx.room.Query
import androidx.room.Update
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.databinding.SavedServerBinding
import io.annaclemens.xivchat.ui.servers.ServersViewModel
import io.sentry.Sentry

@Entity(tableName = "servers")
data class SavedServer(
    @PrimaryKey(autoGenerate = true)
    val id: Int,
    val name: String,
    val host: String? = null,
    val port: Int? = null,
    val auth: String? = null,
    val target: String? = null,
)

@Dao
interface SavedServerDao {
    @Query("select * from servers order by `id` asc")
    fun getAll(): LiveData<List<SavedServer>>

    @Insert(onConflict = OnConflictStrategy.IGNORE)
    suspend fun insert(server: SavedServer)

    @Update(onConflict = OnConflictStrategy.REPLACE)
    suspend fun update(server: SavedServer)

    @Delete
    suspend fun delete(server: SavedServer)

    @Query("delete from servers where `id` = :id")
    suspend fun delete(id: Int)
}

class SavedServerRepository(private val dao: SavedServerDao) {
    val allServers: LiveData<List<SavedServer>> = this.dao.getAll()

    suspend fun insert(server: SavedServer) {
        this.dao.insert(server)
    }

    suspend fun update(server: SavedServer) {
        this.dao.update(server)
    }

    suspend fun delete(server: SavedServer) {
        this.dao.delete(server)
    }

    suspend fun delete(id: Int) = this.dao.delete(id)
}

class SavedServerAdapter internal constructor(context: Context, private val vm: ServersViewModel) : RecyclerView.Adapter<SavedServerAdapter.ViewHolder>() {
    private val inflater = LayoutInflater.from(context)
    private var servers = emptyList<SavedServer>()

    inner class ViewHolder(itemView: View) : RecyclerView.ViewHolder(itemView) {
        private val binding = SavedServerBinding.bind(itemView)

        lateinit var server: SavedServer
        val name: TextView = this.binding.name
        val address: TextView = this.binding.address
        val indicator: ProgressBar = this.binding.connectingIndicator
        val connectedCheck: ImageView = this.binding.connectedCheck
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): ViewHolder {
        val itemView = this.inflater.inflate(R.layout.saved_server, parent, false)
        return this.ViewHolder(itemView)
    }

    override fun onBindViewHolder(holder: ViewHolder, position: Int) {
        val current = this.servers[position]

        fun setCheckVisibility() {
            holder.connectedCheck.visibility = if (this.vm.connected.value == true && this.vm.connectedServerId == current.id) {
                View.VISIBLE
            } else {
                View.GONE
            }
        }

        holder.server = current
        holder.name.text = current.name
        setCheckVisibility()
        holder.itemView.isClickable = !(this.vm.connected.value ?: false)
        if (current.host == null) {
            holder.address.setText(R.string.add_server_type_relay)
        } else {
            @SuppressLint("SetTextI18n")
            holder.address.text = "${current.host}:${current.port}"
        }

        holder.itemView.setOnClickListener {
            val serv = this.vm.connectionService.value ?: return@setOnClickListener

            if (serv.connecting.value == true || this.vm.connected.value == true) {
                return@setOnClickListener
            }

            it.isClickable = false
            holder.indicator.visibility = View.VISIBLE

            val observer = object : Observer<Boolean> {
                private var first = true

                override fun onChanged(connecting: Boolean) {
                    if (this.first) {
                        this.first = false
                        return
                    }

                    if (!connecting) {
                        it.isClickable = true
                        holder.indicator.visibility = View.GONE
                        setCheckVisibility()

                        serv.connecting.removeObserver(this)
                    }
                }
            }
            serv.connecting.observeForever(observer)

            this.vm.connectedServerId = current.id

            if (current.host != null && current.port != null) {
                serv.connect(current.host, current.port)
            } else if (current.auth != null && current.target != null) {
                serv.connectRelay(current.auth, current.target)
            } else {
                // users should not be able to create invalid servers, so capture an event if they do
                Sentry.withScope { scope ->
                    scope.setContexts("server", mapOf(
                        "host" to (current.host ?: "<null>"),
                        "port" to (current.port ?: "<null>"),
                        "auth" to (current.auth ?: "<null>"),
                        "target" to (current.target ?: "<null>")
                    ))

                    Sentry.captureMessage("Attempted connection to invalid server")
                }
            }
        }
    }

    internal fun setServers(servers: List<SavedServer>) {
        this.servers = servers
        this.notifyDataSetChanged()
    }

    override fun getItemCount() = this.servers.size
}
