package io.annaclemens.xivchat.model

import androidx.lifecycle.LiveData
import androidx.room.Dao
import androidx.room.Delete
import androidx.room.Entity
import androidx.room.Ignore
import androidx.room.Insert
import androidx.room.OnConflictStrategy
import androidx.room.PrimaryKey
import androidx.room.Query
import androidx.room.Update
import io.annaclemens.xivchat.ui.FilterType

@Entity(tableName = "tabs")
data class Tab(
    @PrimaryKey(autoGenerate = true)
    val id: Int,
    val name: String,
    val filters: String,
) {
    constructor(id: Int, name: String, filters: Collection<FilterType>) : this(
        id,
        name,
        filters.joinToString(",") { it.name }
    )

    @Ignore
    private val parsedFilters = this.filters()

    fun allowed(code: ChatCode): Boolean = this.parsedFilters.any { it.matches(code) }

    fun filters(): Set<FilterType> = this.filters
        .split(',')
        .mapNotNull {
            try {
                FilterType.valueOf(it)
            } catch (ex: IllegalArgumentException) {
                null
            }
        }
        .toSet()
}

@Dao
interface TabDao {
    @Query("select * from tabs order by `id` asc")
    fun getAll(): LiveData<List<Tab>>

    @Query("select * from tabs where `id` = :id")
    fun get(id: Int): LiveData<Tab>

    @Insert(onConflict = OnConflictStrategy.IGNORE)
    suspend fun insert(tab: Tab)

    @Update(onConflict = OnConflictStrategy.REPLACE)
    suspend fun update(tab: Tab)

    @Delete
    suspend fun delete(tab: Tab)

    @Query("delete from tabs where `id` = :id")
    suspend fun delete(id: Int)

    @Query("select count(*) > 0 from tabs where `id` = :id")
    fun contains(id: Int): Boolean

    @Query("select count(*) from tabs")
    suspend fun count(): Int
}

class TabRepository(private val dao: TabDao) {
    val allTabs: LiveData<List<Tab>> = this.dao.getAll()

    fun get(id: Int): LiveData<Tab> = this.dao.get(id)

    suspend fun insert(tab: Tab) = this.dao.insert(tab)

    suspend fun update(tab: Tab) = this.dao.update(tab)

    suspend fun delete(tab: Tab) = this.dao.delete(tab)

    suspend fun delete(id: Int) = this.dao.delete(id)

    fun contains(id: Int): Boolean = this.dao.contains(id)

    suspend fun addDefaults() {
        if (this.dao.count() > 0) {
            return
        }

        this.insert(Tab(0, "General", FilterType.default()))
        this.insert(Tab(0, "Battle", FilterType.battle()))
    }
}
