package io.annaclemens.xivchat.model.message

import android.content.Context
import io.annaclemens.xivchat.model.ChatCode
import io.annaclemens.xivchat.util.SeString
import io.annaclemens.xivchat.util.rgbaToArgb
import io.noties.markwon.Markwon
import org.msgpack.core.MessagePack
import org.msgpack.core.MessageUnpacker
import java.nio.ByteBuffer
import java.util.*

@OptIn(ExperimentalUnsignedTypes::class)
data class ServerMessage(
    val timestamp: Date,
    val channel: UShort,
    val sender: ByteArray,
    val content: ByteArray,
    val chunks: List<Chunk>,
) {
    fun spans(markdown: Markwon?, context: Context?) = SeString.chunksToSpans(this.chunks, markdown, context)
    fun code(): ChatCode = ChatCode(this.channel.toUInt())

    companion object {
        fun read(input: ByteBuffer): ServerMessage {
            val unpacker = MessagePack.newDefaultUnpacker(input)
            return this.read(unpacker)
        }

        fun read(unpacker: MessageUnpacker): ServerMessage {
            unpacker.unpackArrayHeader()

            val rawTimestamp = unpacker.unpackLong()
            val timestamp = Date(rawTimestamp)

            val channel = unpacker.unpackShort().toUShort()

            val senderLen = unpacker.unpackBinaryHeader()
            val sender = unpacker.readPayload(senderLen)

            val contentLen = unpacker.unpackBinaryHeader()
            val content = unpacker.readPayload(contentLen)

            val chunksLen = unpacker.unpackArrayHeader()
            val chunks = ArrayList<Chunk>(chunksLen)
            for (i in 0 until chunksLen) {
                chunks.add(Chunk.decode(unpacker))
            }

            return ServerMessage(timestamp, channel, sender, content, chunks)
        }
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        @Suppress("ImplicitThis")
        if (this.javaClass != other?.javaClass) return false

        other as ServerMessage

        if (this.timestamp != other.timestamp) return false
        if (this.channel != other.channel) return false
        if (!this.sender.contentEquals(other.sender)) return false
        if (!this.content.contentEquals(other.content)) return false
        if (this.chunks != other.chunks) return false

        return true
    }

    override fun hashCode(): Int {
        var result = this.timestamp.hashCode()
        result = 31 * result + this.channel.hashCode()
        result = 31 * result + this.sender.contentHashCode()
        result = 31 * result + this.content.contentHashCode()
        result = 31 * result + this.chunks.hashCode()
        return result
    }
}

abstract class Chunk {
    companion object {
        @OptIn(ExperimentalUnsignedTypes::class)
        fun decode(unpacker: MessageUnpacker): Chunk {
            unpacker.unpackArrayHeader()
            val tag = unpacker.unpackByte()
            unpacker.unpackArrayHeader()

            return when (tag) {
                ChunkType.Text.code -> TextChunk(
                    fallbackColour = if (unpacker.tryUnpackNil()) null else unpacker.unpackLong().toUInt(),
                    foreground = if (unpacker.tryUnpackNil()) null else unpacker.unpackLong().toUInt(),
                    glow = if (unpacker.tryUnpackNil()) null else unpacker.unpackLong().toUInt(),
                    italic = unpacker.unpackBoolean(),
                    content = unpacker.unpackString(),
                )
                ChunkType.Icon.code -> IconChunk(
                    index = unpacker.unpackByte(),
                )
                else -> throw IllegalArgumentException()
            }
        }
    }
}

enum class ChunkType(val code: Byte) {
    Text(1),
    Icon(2),
}

@OptIn(ExperimentalUnsignedTypes::class)
data class TextChunk(
    val fallbackColour: UInt?,
    val foreground: UInt?,
    val glow: UInt?,
    val italic: Boolean,
    val content: String,
) : Chunk() {
    fun argb(): Long? {
        val colour = this.foreground ?: this.fallbackColour ?: return null
        return rgbaToArgb(colour.toLong())
    }
}

data class IconChunk(val index: Byte) : Chunk()

class Availability(val available: Boolean) {
    companion object {
        fun read(input: ByteBuffer): Availability {
            val unpacker = MessagePack.newDefaultUnpacker(input)

            unpacker.unpackArrayHeader()
            val available = unpacker.unpackBoolean()

            return Availability(available)
        }
    }
}

data class PlayerData(val homeWorld: String, val currentWorld: String, val location: String, val name: String) {
    companion object {
        fun read(input: ByteBuffer): PlayerData? {
            if (!input.hasRemaining()) {
                return null
            }

            val unpacker = MessagePack.newDefaultUnpacker(input)

            unpacker.unpackArrayHeader()

            val homeWorld = unpacker.unpackString()
            val currentWorld = unpacker.unpackString()
            val location = unpacker.unpackString()
            val name = unpacker.unpackString()

            return PlayerData(homeWorld, currentWorld, location, name)
        }
    }
}

data class Channel(val code: Byte, val name: String) {
    companion object {
        fun read(input: ByteBuffer): Channel {
            val unpacker = MessagePack.newDefaultUnpacker(input)

            unpacker.unpackArrayHeader()

            val code = unpacker.unpackByte()
            val name = unpacker.unpackString()

            return Channel(code, name)
        }
    }

    fun channel(): InputChannel? {
        return InputChannel.fromCode(this.code)
    }
}

data class ServerBacklog(val messages: Array<ServerMessage>) {
    companion object {
        fun read(input: ByteBuffer): ServerBacklog {
            val unpacker = MessagePack.newDefaultUnpacker(input)

            unpacker.unpackArrayHeader()

            val numMsgs = unpacker.unpackArrayHeader()
            val msgs = mutableListOf<ServerMessage>()
            for (i in 0 until numMsgs) {
                msgs.add(ServerMessage.read(unpacker))
            }

            return ServerBacklog(msgs.toTypedArray())
        }
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        @Suppress("ImplicitThis")
        if (this.javaClass != other?.javaClass) return false

        other as ServerBacklog

        if (!this.messages.contentEquals(other.messages)) return false

        return true
    }

    override fun hashCode(): Int {
        return this.messages.contentHashCode()
    }
}

data class ServerPlayerList(val type: PlayerListType?, val players: Array<Player>) {
    companion object {
        fun read(input: ByteBuffer): ServerPlayerList {
            val unpacker = MessagePack.newDefaultUnpacker(input)

            unpacker.unpackArrayHeader()

            val type = PlayerListType.fromCode(unpacker.unpackByte())

            val numPlayers = unpacker.unpackArrayHeader()
            val players = mutableListOf<Player>()
            for (i in 0 until numPlayers) {
                players.add(Player.read(unpacker))
            }

            return ServerPlayerList(type, players.toTypedArray())
        }
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        @Suppress("ImplicitThis")
        if (this.javaClass != other?.javaClass) return false

        other as ServerPlayerList

        if (!this.players.contentEquals(other.players)) return false

        return true
    }

    override fun hashCode(): Int {
        return this.players.contentHashCode()
    }

}

@OptIn(ExperimentalUnsignedTypes::class)
data class Player(
    val name: String?,
    val fc: String?,
    val status: ULong,
    val currentWorld: UShort,
    val currentWorldName: String?,
    val homeWorld: UShort,
    val homeWorldName: String?,
    val territory: UShort,
    val territoryName: String?,
    val job: Byte,
    val jobName: String?,
    val grandCompany: Byte,
    val grandCompanyName: String?,
    val languages: Byte,
    val mainLanguage: Byte,
) {
    companion object {
        fun read(unpacker: MessageUnpacker): Player {
            unpacker.unpackArrayHeader()

            return Player(
                name = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                fc = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                status = unpacker.unpackLong().toULong(),
                currentWorld = unpacker.unpackShort().toUShort(),
                currentWorldName = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                homeWorld = unpacker.unpackShort().toUShort(),
                homeWorldName = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                territory = unpacker.unpackShort().toUShort(),
                territoryName = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                job = unpacker.unpackByte(),
                jobName = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                grandCompany = unpacker.unpackByte(),
                grandCompanyName = if (unpacker.tryUnpackNil()) null else unpacker.unpackString(),
                languages = unpacker.unpackByte(),
                mainLanguage = unpacker.unpackShort().toByte(),
            )
        }
    }

    private fun rawStatus(shift: Int) = (this.status and (1.toULong() shl shift)) > 0u
    fun status(status: OnlineStatus) = this.rawStatus(status.code.toInt())
}
