package io.annaclemens.xivchat.ui

import android.os.Bundle
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.CompoundButton
import android.widget.LinearLayout
import androidx.activity.viewModels
import androidx.appcompat.widget.SwitchCompat
import androidx.core.content.ContextCompat
import androidx.fragment.app.DialogFragment
import androidx.fragment.app.FragmentManager
import androidx.lifecycle.MutableLiveData
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.databinding.FilterSelectionBinding
import io.annaclemens.xivchat.model.ChatCode
import io.annaclemens.xivchat.model.ChatSource
import io.annaclemens.xivchat.model.ChatType
import io.annaclemens.xivchat.model.Tab
import io.annaclemens.xivchat.ui.messages.MessagesViewModel
import io.annaclemens.xivchat.util.getApp

class FilterSelection : DialogFragment() {
    companion object {
        const val ARG_TAB_ID = "tabId"
        const val ARG_TAB_NEW = "tabNew"
        const val ARG_TAB_NEW_NAME = "tabNewName"

        fun openExisting(manager: FragmentManager, id: Int) {
            val fragment = FilterSelection()
            fragment.arguments = Bundle().apply {
                this.putInt(this@Companion.ARG_TAB_ID, id)
            }

            fragment.show(manager, "filter_selection")
        }

        fun openNew(manager: FragmentManager, name: String) {
            val fragment = FilterSelection()
            fragment.arguments = Bundle().apply {
                this.putBoolean(this@Companion.ARG_TAB_NEW, true)
                this.putString(this@Companion.ARG_TAB_NEW_NAME, name)
            }

            fragment.show(manager, "filter_selection")
        }
    }

    private lateinit var binding: FilterSelectionBinding
    private val tab: MutableLiveData<Tab> = MutableLiveData()
    private lateinit var filters: TabFilter

    private fun saveFilters() {
        val oldTab = this.tab.value ?: return

        val tab = Tab(
            oldTab.id,
            oldTab.name,
            this.filters.set.joinToString(",") { it.name },
        )

        val vm: MessagesViewModel by this.getApp().viewModels()
        if (tab.id == 0) {
            vm.insertTab(tab)
        } else {
            vm.updateTab(tab)
        }
    }

    override fun onStart() {
        super.onStart()
        this.dialog?.window?.setLayout(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.MATCH_PARENT)
    }

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        val vm: MessagesViewModel by this.getApp().viewModels()
        this.binding = FilterSelectionBinding.inflate(inflater)

        // finish set up once we receive the tab from the database
        this.tab.observe(this.viewLifecycleOwner) {
            this.finishSetup(it)
        }

        // check if we're being asked to configure a new tab or an existing tab
        val newTab = this.requireArguments().getBoolean(Companion.ARG_TAB_NEW)
        val tabId = this.requireArguments().getInt(Companion.ARG_TAB_ID, 0)
        if (newTab) {
            // set the tab immediately if it's a new tab
            val tabName = this.requireArguments().getString(Companion.ARG_TAB_NEW_NAME)!!
            this.tab.value = Tab(0, tabName, FilterType.default().joinToString(",") { it.name })
        } else {
            // download the tab from the database if it's an existing tab
            vm.getTab(tabId).observe(this.viewLifecycleOwner) {
                this.tab.value = it
            }
        }

        return this.binding.root
    }

    private fun finishSetup(tab: Tab) {
        fun divider() = View(this.context, null, android.R.attr.listDivider).apply {
            this.layoutParams = LinearLayout.LayoutParams(
                LinearLayout.LayoutParams.MATCH_PARENT,
                this.context.resources.displayMetrics.density.toInt(),
            )
            this.setBackgroundColor(ContextCompat.getColor(this@FilterSelection.requireContext(), R.color.colorAccent))
        }

        fun space() = View(this.context).apply {
            this.layoutParams = LinearLayout.LayoutParams(
                LinearLayout.LayoutParams.MATCH_PARENT,
                16 * this.context.resources.displayMetrics.density.toInt(),
            )
        }

        // at this point, we've received the tab, so we're no longer loading
        this.binding.progressBar.visibility = View.GONE

        // set up filters
        this.filters = TabFilter(tab.filters())

        // get all the categories and make a map of their switches
        val categories = FilterCategory.values()
        val switches: MutableMap<FilterCategory, MutableList<SwitchCompat>> = mutableMapOf()

        for (entry in categories.withIndex()) {
            val category = entry.value

            // create a master switch for the category
            val categorySwitch = SwitchCompat(this.requireContext()).apply {
                this.setText(category.stringId)
                this.textSize = 16f
                this.layoutParams = LinearLayout.LayoutParams(
                    LinearLayout.LayoutParams.MATCH_PARENT,
                    LinearLayout.LayoutParams.WRAP_CONTENT,
                )
            }

            // create the checked listener for the master switch, checking/unchecking all child switches
            val categoryCheckedListener = CompoundButton.OnCheckedChangeListener { _, isChecked ->
                switches[category]?.forEach {
                    it.isChecked = isChecked
                }
            }

            this.binding.linearLayout.addView(categorySwitch)
            this.binding.linearLayout.addView(divider())

            for (type in category.types) {
                // create a switch for the individual type
                val typeSwitch = SwitchCompat(this.requireContext()).apply {
                    this.layoutParams = LinearLayout.LayoutParams(
                        LinearLayout.LayoutParams.MATCH_PARENT,
                        LinearLayout.LayoutParams.WRAP_CONTENT,
                    )
                    this.setText(type.stringId)
                    this.isChecked = this@FilterSelection.filters.contains(type)

                    this.setOnCheckedChangeListener { _, isChecked ->
                        // check if all switches are checked now
                        val allChecked = switches[category]?.all { it.isChecked } ?: false
                        // turn off the category listener temporarily
                        categorySwitch.setOnCheckedChangeListener(null)
                        // handle the category switch and modifying the filters
                        if (isChecked) {
                            if (allChecked) {
                                categorySwitch.isChecked = true
                            }
                            this@FilterSelection.filters.add(type)
                        } else {
                            if (!allChecked) {
                                categorySwitch.isChecked = false
                            }
                            this@FilterSelection.filters.remove(type)
                        }
                        // restore the category listener
                        categorySwitch.setOnCheckedChangeListener(categoryCheckedListener)
                    }
                }

                // add type switch to category list
                if (switches.containsKey(category)) {
                    switches[category]!!.add(typeSwitch)
                } else {
                    switches[category] = mutableListOf(typeSwitch)
                }

                this.binding.linearLayout.addView(typeSwitch)
            }

            // set if the category switch is checked based on if all its children are checked
            categorySwitch.isChecked = switches[category]?.all { it.isChecked } ?: false

            // set the listener finally
            categorySwitch.setOnCheckedChangeListener(categoryCheckedListener)

            // only add space after the category if it's not the last category
            if (entry.index != categories.lastIndex) {
                this.binding.linearLayout.addView(space())
            }
        }

        // set up buttons and text
        this.binding.headerText.visibility = View.VISIBLE
        this.binding.buttonLayout.visibility = View.VISIBLE
        this.binding.okButton.setOnClickListener {
            this.saveFilters()
            this.dismiss()
        }
        this.binding.cancelButton.setOnClickListener {
            this.dismiss()
        }
    }
}

enum class FilterCategory(val stringId: Int, val types: Collection<FilterType>) {
    Chat(R.string.chat_category_chat, listOf(
        FilterType.Say,
        FilterType.Yell,
        FilterType.Shout,
        FilterType.Tell,
        FilterType.Party,
        FilterType.Alliance,
        FilterType.FreeCompany,
        FilterType.PvpTeam,
        FilterType.CrossWorldLinkshell1,
        FilterType.CrossWorldLinkshell2,
        FilterType.CrossWorldLinkshell3,
        FilterType.CrossWorldLinkshell4,
        FilterType.CrossWorldLinkshell5,
        FilterType.CrossWorldLinkshell6,
        FilterType.CrossWorldLinkshell7,
        FilterType.CrossWorldLinkshell8,
        FilterType.Linkshell1,
        FilterType.Linkshell2,
        FilterType.Linkshell3,
        FilterType.Linkshell4,
        FilterType.Linkshell5,
        FilterType.Linkshell6,
        FilterType.Linkshell7,
        FilterType.Linkshell8,
        FilterType.NoviceNetwork,
        FilterType.StandardEmote,
        FilterType.CustomEmote,
        FilterType.Gm,
    )),
    Battle(R.string.chat_category_battle, listOf(
        FilterType.Battle,
    )),
    Announcements(R.string.chat_category_announcements, listOf(
        FilterType.Debug,
        FilterType.Urgent,
        FilterType.Notice,
        FilterType.SystemMessages,
        FilterType.OwnBattleSystem,
        FilterType.OthersBattleSystem,
        FilterType.GatheringSystem,
        FilterType.ErrorMessages,
        FilterType.Echo,
        FilterType.NoviceNetworkAnnouncements,
        FilterType.FreeCompanyAnnouncements,
        FilterType.PvpTeamAnnouncements,
        FilterType.FreeCompanyLoginLogout,
        FilterType.PvpTeamLoginLogout,
        FilterType.RetainerSale,
        FilterType.NpcDialogue,
        FilterType.NpcAnnouncement,
        FilterType.Loot,
        FilterType.OwnProgression,
        FilterType.PartyProgression,
        FilterType.OthersProgression,
        FilterType.OwnLoot,
        FilterType.OthersLoot,
        FilterType.OwnCrafting,
        FilterType.OthersCrafting,
        FilterType.OwnGathering,
        FilterType.OthersFishing,
        FilterType.PeriodicRecruitment,
        FilterType.Sign,
        FilterType.Random,
        FilterType.Orchestrion,
        FilterType.MessageBook,
        FilterType.Alarm,
    )),
}

enum class FilterType(val stringId: Int, private val types: Collection<ChatType>, private val sources: Collection<ChatSource> = emptyList()) {
    Say(R.string.chat_say, ChatType.Say),
    Shout(R.string.chat_shout, ChatType.Shout),
    Yell(R.string.chat_yell, ChatType.Yell),
    Tell(R.string.chat_tell, listOf(
        ChatType.TellOutgoing,
        ChatType.TellIncoming,
    )),
    Party(R.string.chat_party, listOf(
        ChatType.Party,
        ChatType.CrossParty,
    )),
    Alliance(R.string.chat_alliance, ChatType.Alliance),
    FreeCompany(R.string.chat_free_company, ChatType.FreeCompany),
    PvpTeam(R.string.chat_pvp_team, ChatType.PvpTeam),
    CrossWorldLinkshell1(R.string.chat_cross_world_linkshell_1, ChatType.CrossLinkshell1),
    CrossWorldLinkshell2(R.string.chat_cross_world_linkshell_2, ChatType.CrossLinkshell2),
    CrossWorldLinkshell3(R.string.chat_cross_world_linkshell_3, ChatType.CrossLinkshell3),
    CrossWorldLinkshell4(R.string.chat_cross_world_linkshell_4, ChatType.CrossLinkshell4),
    CrossWorldLinkshell5(R.string.chat_cross_world_linkshell_5, ChatType.CrossLinkshell5),
    CrossWorldLinkshell6(R.string.chat_cross_world_linkshell_6, ChatType.CrossLinkshell6),
    CrossWorldLinkshell7(R.string.chat_cross_world_linkshell_7, ChatType.CrossLinkshell7),
    CrossWorldLinkshell8(R.string.chat_cross_world_linkshell_8, ChatType.CrossLinkshell8),
    Linkshell1(R.string.chat_linkshell_1, ChatType.Linkshell1),
    Linkshell2(R.string.chat_linkshell_2, ChatType.Linkshell2),
    Linkshell3(R.string.chat_linkshell_3, ChatType.Linkshell3),
    Linkshell4(R.string.chat_linkshell_4, ChatType.Linkshell4),
    Linkshell5(R.string.chat_linkshell_5, ChatType.Linkshell5),
    Linkshell6(R.string.chat_linkshell_6, ChatType.Linkshell6),
    Linkshell7(R.string.chat_linkshell_7, ChatType.Linkshell7),
    Linkshell8(R.string.chat_linkshell_8, ChatType.Linkshell8),
    NoviceNetwork(R.string.chat_novice_network, ChatType.NoviceNetwork),
    StandardEmote(R.string.chat_standard_emotes, ChatType.StandardEmote),
    CustomEmote(R.string.chat_custom_emotes, ChatType.CustomEmote),

    Battle(R.string.chat_battle, listOf(
        ChatType.Damage,
        ChatType.Miss,
        ChatType.Action,
        ChatType.Item,
        ChatType.Healing,
        ChatType.GainBuff,
        ChatType.LoseBuff,
        ChatType.GainDebuff,
        ChatType.LoseDebuff,
        ChatType.BattleSystem,
    )),

    Debug(R.string.chat_debug, ChatType.Debug),
    Urgent(R.string.chat_urgent, ChatType.Urgent),
    Notice(R.string.chat_notice, ChatType.Notice),
    SystemMessages(R.string.chat_system_messages, ChatType.System),
    OwnBattleSystem(R.string.chat_own_battle_system_messages, ChatType.BattleSystem, ChatSource.Self),
    OthersBattleSystem(R.string.chat_others_battle_system_messages, ChatType.BattleSystem, ChatSource.OTHERS),
    GatheringSystem(R.string.chat_gathering_system_messages, ChatType.GatheringSystem),
    ErrorMessages(R.string.chat_error_messages, ChatType.Error),
    Echo(R.string.chat_echo, ChatType.Echo),
    NoviceNetworkAnnouncements(R.string.chat_novice_network_notifications, ChatType.NoviceNetworkSystem),
    FreeCompanyAnnouncements(R.string.chat_free_company_announcements, ChatType.FreeCompanyAnnouncement),
    PvpTeamAnnouncements(R.string.chat_pvp_team_announcements, ChatType.PvpTeamAnnouncement),
    FreeCompanyLoginLogout(R.string.chat_free_company_member_login_notifications, ChatType.FreeCompanyLoginLogout),
    PvpTeamLoginLogout(R.string.chat_pvp_team_member_login_notifications, ChatType.PvpTeamLoginLogout),
    RetainerSale(R.string.chat_retainer_sale_notifications, ChatType.RetainerSale),
    NpcDialogue(R.string.chat_npc_dialogue, ChatType.NpcDialogue),
    NpcAnnouncement(R.string.chat_npc_dialogue_announcements, ChatType.NpcAnnouncement),
    Loot(R.string.chat_loot_notices, ChatType.LootNotice),
    OwnProgression(R.string.chat_own_progression_messages, ChatType.Progress, ChatSource.Self),
    PartyProgression(R.string.chat_party_members_progression_messages, ChatType.Progress, ChatSource.PartyMember),
    OthersProgression(R.string.chat_others_progression_messages, ChatType.Progress, ChatSource.Other),
    OwnLoot(R.string.chat_own_loot_messages, ChatType.LootRoll, ChatSource.Self),
    OthersLoot(R.string.chat_others_loot_messages, ChatType.LootRoll, ChatSource.OTHERS),
    OwnCrafting(R.string.chat_own_synthesis_messages, ChatType.Crafting, ChatSource.Self),
    OthersCrafting(R.string.chat_others_synthesis_messages, ChatType.Crafting, ChatSource.OTHERS),
    OwnGathering(R.string.chat_own_gathering_messages, ChatType.Gathering, ChatSource.Self),
    OthersFishing(R.string.chat_others_fishing_messages, ChatType.Gathering, ChatSource.OTHERS),
    PeriodicRecruitment(R.string.chat_periodic_recruitment_notifications, ChatType.PeriodicRecruitmentNotification),
    Sign(R.string.chat_sign_messages_for_pc_targets, ChatType.Sign),
    Random(R.string.chat_random_number_messages, ChatType.RandomNumber),
    Orchestrion(R.string.chat_current_orchestrion_track_messages, ChatType.Orchestrion),
    MessageBook(R.string.chat_message_book_alert, ChatType.MessageBook),
    Alarm(R.string.chat_alarm_notifications, ChatType.Alarm),

    Gm(R.string.chat_gm, listOf(
        ChatType.GmTell,
        ChatType.GmSay,
        ChatType.GmShout,
        ChatType.GmYell,
        ChatType.GmParty,
        ChatType.GmFreeCompany,
        ChatType.GmLinkshell1,
        ChatType.GmLinkshell2,
        ChatType.GmLinkshell3,
        ChatType.GmLinkshell4,
        ChatType.GmLinkshell5,
        ChatType.GmLinkshell6,
        ChatType.GmLinkshell7,
        ChatType.GmLinkshell8,
        ChatType.GmNoviceNetwork,
    ));

    companion object {
        fun default(): Set<FilterType> {
            return FilterCategory.Chat.types
                .asSequence()
                .plus(FilterCategory.Announcements.types)
                .minus(OwnBattleSystem)
                .minus(OthersBattleSystem)
                .minus(NpcDialogue)
                .minus(OthersFishing)
                .toSet()
        }

        fun battle(): Set<FilterType> {
            return FilterCategory.Battle.types
                .asSequence()
                .plus(OwnBattleSystem)
                .plus(OthersBattleSystem)
                .toSet()
        }
    }

    constructor(stringId: Int, type: ChatType, sources: Collection<ChatSource> = emptyList()) : this(stringId, listOf(type), sources)
    constructor(stringId: Int, type: ChatType, source: ChatSource) : this(stringId, listOf(type), listOf(source))

    fun matches(code: ChatCode): Boolean {
        return this.types.contains(code.type()) && (this.sources.isEmpty() || this.sources.contains(code.source()))
    }
}

class TabFilter(types: Collection<FilterType>) {
    private val types: MutableSet<FilterType> = types.toMutableSet()

    val set: Set<FilterType>
        get() = this.types

    fun allowed(type: ChatCode): Boolean {
        return this.types.any { it.matches(type) }
    }

    fun contains(type: FilterType): Boolean {
        return this.types.contains(type)
    }

    fun add(type: FilterType) {
        this.types.add(type)
    }

    fun remove(type: FilterType) {
        this.types.remove(type)
    }
}
