package io.annaclemens.xivchat.ui.friends

import android.content.Context
import android.content.res.ColorStateList
import android.view.LayoutInflater
import android.view.MenuInflater
import android.view.View
import android.view.ViewGroup
import android.widget.ImageView
import android.widget.TextView
import androidx.core.content.ContextCompat
import androidx.core.widget.ImageViewCompat
import androidx.navigation.Navigation
import androidx.navigation.findNavController
import androidx.recyclerview.widget.RecyclerView
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.databinding.PlayerListElementBinding
import io.annaclemens.xivchat.model.message.OnlineStatus
import io.annaclemens.xivchat.model.message.Player

class FriendListAdapter(private val context: Context) : RecyclerView.Adapter<FriendListAdapter.ViewHolder>() {
    inner class ViewHolder(itemView: View) : RecyclerView.ViewHolder(itemView) {
        private val binding = PlayerListElementBinding.bind(itemView)

        val statusIcon: ImageView = this.binding.statusIcon
        val playerName: TextView = this.binding.name
        val freeCompanyName: TextView = this.binding.freeCompanyName
        val playerCurrentWorld: TextView = this.binding.playerLocation
    }

    private val inflater = LayoutInflater.from(this.context)
    private var playerList: List<Player> = emptyList()

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): ViewHolder {
        val itemView = this.inflater.inflate(R.layout.player_list_element, parent, false)
        return this.ViewHolder(itemView)
    }

    @OptIn(ExperimentalUnsignedTypes::class)
    override fun onBindViewHolder(holder: ViewHolder, position: Int) {
        val player = this.playerList[position]

        if (player.name != null) {
            holder.itemView.setOnClickListener(Navigation.createNavigateOnClickListener(FriendListFragmentDirections.actionToPlayerDetail(
                playerName = player.name,
                playerWorld = player.homeWorld.toInt(),
            )))

            if (player.homeWorldName != null) {
                holder.itemView.setOnCreateContextMenuListener { menu, _, _ ->
                    MenuInflater(this.context).inflate(R.menu.friend_entry_menu, menu)

                    menu.findItem(R.id.send_tell).setOnMenuItemClickListener {
                        val action = FriendListFragmentDirections.actionSendTell(player.name, player.homeWorldName)
                        holder.itemView.findNavController().navigate(action)
                        true
                    }
                }
            }
        }

        holder.playerName.text = player.name ?: this.context.getString(R.string.friend_list_could_not_retrieve)
        holder.freeCompanyName.text = player.fc?.let { "«${it}»" } ?: ""

        var currentWorldText = player.currentWorldName
        if (player.territoryName != null) {
            currentWorldText += " / ${player.territoryName}"
        }
        holder.playerCurrentWorld.text = currentWorldText

        // determine icon to use, its colour, and its alt text
        val resources = when {
            player.status(OnlineStatus.SharingDuty) -> Triple(R.drawable.app_status_contents, R.color.statusOnline, R.string.status_sharing_duty)
            player.status(OnlineStatus.SimilarDuty) -> Triple(R.drawable.app_status_contents, R.color.statusSimilarDuty, R.string.status_similar_duty)
            player.status(OnlineStatus.AnotherWorld) -> Triple(R.drawable.app_status_contents, R.color.statusAfk, R.string.status_another_world)
            player.status(OnlineStatus.AwayFromKeyboard) -> Triple(R.drawable.app_status_afk, R.color.statusAfk, R.string.status_away_from_keyboard)
            player.status(OnlineStatus.RolePlaying) -> Triple(R.drawable.app_status_roleplaying, R.color.statusRoleplaying, R.string.status_role_playing)
            player.status(OnlineStatus.PartyLeaderCrossWorld) -> Triple(R.drawable.app_status_party_leader, R.color.statusPartyCross, R.string.status_party_leader_cross_world)
            player.status(OnlineStatus.PartyMemberCrossWorld) -> Triple(R.drawable.app_status_party_member, R.color.statusPartyCross, R.string.status_party_member_cross_world)
            player.status(OnlineStatus.PartyLeader) -> Triple(R.drawable.app_status_party_leader, R.color.statusParty, R.string.status_party_leader)
            player.status(OnlineStatus.PartyMember) -> Triple(R.drawable.app_status_party_member, R.color.statusParty, R.string.status_party_member)
            player.status(OnlineStatus.Online) -> Triple(R.drawable.app_status_online, R.color.statusOnline, R.string.status_online)
            else -> Triple(R.drawable.app_status_online, R.color.statusOffline, R.string.status_offline)
        }
        holder.statusIcon.setImageResource(resources.first)
        holder.statusIcon.contentDescription = this.context.getString(resources.third)
        val colour = ContextCompat.getColor(this.context, resources.second)
        ImageViewCompat.setImageTintList(holder.statusIcon, ColorStateList.valueOf(colour))
    }

    override fun getItemCount() = this.playerList.size

    internal fun setPlayerList(list: List<Player>) {
        this.playerList = list.sortedByDescending {
            it.status(OnlineStatus.Online)
        }
    }
}
