package io.annaclemens.xivchat.ui.friends

import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.net.Uri
import android.os.Bundle
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.TextView
import androidx.activity.viewModels
import androidx.core.view.isEmpty
import androidx.fragment.app.Fragment
import androidx.lifecycle.lifecycleScope
import androidx.navigation.fragment.navArgs
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.databinding.PlayerDetailsBinding
import io.annaclemens.xivchat.model.message.OnlineStatus
import io.annaclemens.xivchat.model.message.Player
import io.annaclemens.xivchat.util.CharacterSearchResults
import io.annaclemens.xivchat.util.getApp
import io.ktor.client.*
import io.ktor.client.engine.android.*
import io.ktor.client.features.*
import io.ktor.client.request.*
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.launch
import kotlinx.serialization.decodeFromString
import kotlinx.serialization.json.Json
import java.util.*

class PlayerDetailFragment : Fragment() {
    @OptIn(ExperimentalUnsignedTypes::class)
    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        val vm: FriendListViewModel by this.getApp().viewModels()
        val binding = PlayerDetailsBinding.inflate(inflater)

        val args: PlayerDetailFragmentArgs by this.navArgs()
        val playerWorld = args.playerWorld.toUShort()
        val player = vm.friends.value?.find { it.name == args.playerName && it.homeWorld == playerWorld }!!

        this.lifecycleScope.launchWhenCreated {
            val cacheKey = "${player.name}/${player.homeWorld}"

            val bitmap = vm.portraitCache[cacheKey] ?: (this@PlayerDetailFragment.downloadPlayerPortrait(player) ?: return@launchWhenCreated)

            this@PlayerDetailFragment.activity?.runOnUiThread {
                vm.portraitCache[cacheKey] = bitmap
                binding.portraitPlaceholder.visibility = View.GONE
                binding.portrait.apply {
                    this.setImageBitmap(bitmap)
                    this.visibility = View.VISIBLE
                }
            }
        }

        binding.name.text = player.name
        binding.homeWorld.text = player.homeWorldName

        binding.freeCompany.text = if (player.fc.isNullOrBlank()) this.getString(R.string.player_detail_no_fc) else player.fc
        binding.grandCompany.text = player.grandCompanyName

        binding.location.visibility = if (player.territoryName.isNullOrBlank()) View.GONE else View.VISIBLE
        binding.location.text = player.territoryName
        binding.currentWorld.text = player.currentWorldName

        binding.jobCard.visibility = if (player.jobName.isNullOrBlank()) View.GONE else View.VISIBLE
        binding.job.text = player.jobName?.let { jobName ->
            jobName
                .split(' ')
                .joinToString(" ") { word ->
                    word.replaceFirstChar { if (it.isLowerCase()) it.titlecase(Locale.getDefault()) else it.toString() }
                }
        }

        fun createStatusText(textRes: Int): TextView {
            return TextView(this.context, null, androidx.appcompat.R.style.TextAppearance_AppCompat_Body1).apply {
                this.setText(textRes)
                this.textSize = 16f
            }
        }

        for (status in OnlineStatus.values()) {
            if (!player.status(status)) {
                continue
            }

            val textView = createStatusText(status.stringRes)
            binding.statusList.addView(textView)
        }

        if (binding.statusList.isEmpty()) {
            binding.statusList.addView(createStatusText(OnlineStatus.Offline.stringRes))
        }

        return binding.root
    }

    private val searchResultsJson = Json {
        this.ignoreUnknownKeys = true
    }

    private suspend fun downloadPlayerPortrait(player: Player): Bitmap? {
        // create the http client and build the xivapi url for searching for this player
        val client = HttpClient(Android) {
            this.install(HttpTimeout.Feature) {
                this.requestTimeoutMillis = 10_000
                this.connectTimeoutMillis = 10_000
                this.socketTimeoutMillis = 10_000
            }
        }
        val url = Uri.parse("https://xivapi.com/character/search")
            .buildUpon()
            .appendQueryParameter("name", player.name)
            .appendQueryParameter("server", player.homeWorldName)
            .appendQueryParameter("columns", "Avatar")
            .build()

        val imageChannel = Channel<Bitmap?>(1)

        // run this on the io dispatcher
        CoroutineScope(Dispatchers.IO).launch {
            val image: Bitmap?
            try {
                // download the search results and decode them into a utf-8 string
                val jsonBytes: ByteArray = client.get(url.toString())
                val json = jsonBytes.decodeToString()
                // parse the search results into a class
                val data: CharacterSearchResults = this@PlayerDetailFragment.searchResultsJson.decodeFromString(json)
                // if there were no results, return false (will try again on next zone change)
                if (data.results.isEmpty()) {
                    imageChannel.send(null)
                    return@launch
                }
                // otherwise, get the avatar url from the first result and download the image bytes
                val imageBytes: ByteArray = client.get(data.results[0].avatar)
                // decode the image into a bitmap for android
                image = BitmapFactory.decodeByteArray(imageBytes, 0, imageBytes.size)
            } catch (ex: Exception) {
                // if any error occurred, log it and try again on next zone change
                ex.printStackTrace()
                imageChannel.send(null)
                return@launch
            } finally {
                // make sure we always close the client
                client.close()
            }

            // if the image got downloaded and parsed and the app is set, tell the app to update the picture
            imageChannel.send(image)
        }

        return imageChannel.receive()
    }
}
