package io.annaclemens.xivchat.ui.messages

import android.os.Bundle
import android.text.SpannableStringBuilder
import android.view.Gravity
import android.view.ViewGroup
import android.widget.TextView
import androidx.core.content.res.ResourcesCompat
import androidx.fragment.app.Fragment
import androidx.lifecycle.Observer
import androidx.recyclerview.widget.LinearLayoutManager
import androidx.recyclerview.widget.RecyclerView
import io.annaclemens.xivchat.MainActivity
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.model.Tab
import io.annaclemens.xivchat.model.message.ServerMessage
import io.annaclemens.xivchat.util.AlertDialogFragment
import io.annaclemens.xivchat.util.MessageFilter
import io.ktor.http.*
import io.noties.markwon.Markwon
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch
import me.saket.bettermovementmethod.BetterLinkMovementMethod
import java.time.ZoneId
import java.time.format.DateTimeFormatter
import java.time.format.FormatStyle

class MessagesAdapter(
    private val app: MainActivity,
    private val fragment: Fragment,
    private val model: MessagesViewModel,
    private val viewManager: RecyclerView.LayoutManager,
    private val markdown: Markwon,
    private val tab: Tab,
) : RecyclerView.Adapter<MessagesAdapter.ViewHolder>() {
    class ViewHolder(val textView: TextView) : RecyclerView.ViewHolder(textView)

    @Volatile
    private var lastIndex = 0

    val filter = MessageFilter(this.tab.filters())
    private val filteredMessages: MutableList<ServerMessage> = ArrayList()

    private val observer = MessageObserver(this)

    private class MessageObserver(val adapter: MessagesAdapter) : Observer<Collection<ServerMessage>> {
        override fun onChanged(it: Collection<ServerMessage>?) {
            if (it == null) {
                return
            }

            val currentMessages = this.adapter.model.messages
            val newMessages = currentMessages.slice(this.adapter.lastIndex until currentMessages.size)

            var amount = 0
            for (added in newMessages) {
                if (this.adapter.filter.allowed(added)) {
                    this.adapter.filteredMessages.add(added)
                    amount += 1
                }
            }
            this.adapter.lastIndex += newMessages.size
            this.adapter.notifyItemRangeInserted(0, amount)
        }
    }

    private val movementMethod = BetterLinkMovementMethod.newInstance().apply {
        this.setOnLinkClickListener { _, url ->
            // check for phishing links - no good link should contain ".com-"
            if (!Url(url).host.contains(".com-")) {
                return@setOnLinkClickListener false
            }

            val desc = this@MessagesAdapter.fragment.getString(R.string.phishing_warning_desc)
            AlertDialogFragment.newInstance(MessagesFragment.DIALOG_PHISHING, Bundle().apply {
                this.putString("message", "${desc}\n\n${url}")
                this.putString("url", url)
            }).show(this@MessagesAdapter.fragment.childFragmentManager, "phishing_warning")

            true
        }
    }

    init {
        this.registerAdapterDataObserver(object : RecyclerView.AdapterDataObserver() {
            override fun onItemRangeInserted(positionStart: Int, itemCount: Int) {
                val vm = this@MessagesAdapter.viewManager
                if (vm is LinearLayoutManager && vm.findFirstVisibleItemPosition() == 0) {
                    this@MessagesAdapter.viewManager.scrollToPosition(0)
                }
            }
        })

        // listen for new messages
        this.registerObserver()
    }

    fun registerObserver() {
        this.model.lastAdded.observe(this.fragment.viewLifecycleOwner, this.observer)
    }

    fun unregisterObserver() {
        this.model.lastAdded.removeObserver(this.observer)
    }

    fun catchUp(): Job {
        val currentMessages = this.model.messages
        val newMessages = currentMessages.slice(this.lastIndex until currentMessages.size)
        return CoroutineScope(Dispatchers.Default).launch {
            val chunks = newMessages.chunked(100).asReversed()
            val insertAt = this@MessagesAdapter.filteredMessages.size
            var notifyIdx = 0
            for (chunk in chunks) {
                val filtered = chunk.filter { this@MessagesAdapter.filter.allowed(it) }
                this@MessagesAdapter.lastIndex += chunk.size
                if (filtered.isEmpty()) {
                    continue
                }
                val notify = if (notifyIdx == 0) 0 else (notifyIdx - 1)
                this@MessagesAdapter.app.runOnUiThread {
                    this@MessagesAdapter.filteredMessages.addAll(insertAt, filtered)
                    this@MessagesAdapter.notifyItemRangeInserted(notify, filtered.size)
                }
                notifyIdx += filtered.size
            }
        }
    }

    fun filterExisting() {
        val len = this.filteredMessages.size
        this.filteredMessages.clear()
        this.notifyItemRangeRemoved(0, len)

        this.lastIndex = 0

        this.catchUp()
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): ViewHolder {
        val textView = TextView(parent.context).apply {
            // allow text to be selected
            this.setTextIsSelectable(true)
            // necessary to allow links to be clicked
            this.movementMethod = this@MessagesAdapter.movementMethod
            // ffxiv special symbols
            this.typeface = ResourcesCompat.getFont(parent.context, R.font.ffxiv)
            // add some bottom padding because letters like "g" get cut off
            // TODO: figure out a better solution
            this.setPadding(0, 0, 0, this.lineHeight / 4)
            this.gravity = Gravity.CENTER_VERTICAL
        }
        return ViewHolder(textView)
    }

    override fun onBindViewHolder(holder: ViewHolder, position: Int) {
        if (this.filteredMessages.isEmpty()) {
            return
        }
        val message = this.filteredMessages.getOrNull(this.filteredMessages.size - position - 1) ?: return
        holder.textView.text = this.processMessage(message)
    }

    private fun processMessage(msg: ServerMessage): CharSequence {
        val dt = msg.timestamp.toInstant().atZone(ZoneId.systemDefault()).toLocalDateTime()
        val time = dt.format(DateTimeFormatter.ofLocalizedTime(FormatStyle.SHORT))

        return SpannableStringBuilder()
            .append('[')
            .append(time)
            .append(']')
            .apply {
                val markdownMode = this@MessagesAdapter.app.markdownMode
                val content = msg.spans(if (markdownMode) this@MessagesAdapter.markdown else null, this@MessagesAdapter.app)
                this.append(content)
            }
    }

    override fun getItemCount() = this.filteredMessages.size
}
