package io.annaclemens.xivchat.ui.messages

import android.annotation.SuppressLint
import android.content.Context
import android.os.Bundle
import android.text.InputType
import android.view.LayoutInflater
import android.view.Menu
import android.view.MenuInflater
import android.view.MenuItem
import android.view.View
import android.view.ViewGroup
import android.view.inputmethod.EditorInfo
import android.view.inputmethod.InputMethodManager
import android.widget.SearchView
import androidx.activity.viewModels
import androidx.core.view.iterator
import androidx.core.widget.addTextChangedListener
import androidx.fragment.app.Fragment
import androidx.lifecycle.ViewModelProvider
import androidx.lifecycle.lifecycleScope
import androidx.navigation.fragment.navArgs
import androidx.viewpager2.adapter.FragmentStateAdapter
import androidx.viewpager2.widget.ViewPager2
import com.google.android.material.tabs.TabLayoutMediator
import io.annaclemens.xivchat.MainActivity
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.databinding.FragmentMessagesTabsBinding
import io.annaclemens.xivchat.model.Tab
import io.annaclemens.xivchat.model.message.ClientMessage
import io.annaclemens.xivchat.ui.FilterSelection
import io.annaclemens.xivchat.ui.servers.ServersViewModel
import io.annaclemens.xivchat.util.AlertDialogFragment
import io.annaclemens.xivchat.util.EditTextDialog
import io.annaclemens.xivchat.util.getApp
import kotlinx.coroutines.launch


class MessagesTabsFragment : Fragment() {
    companion object {
        const val DIALOG_DELETE_TAB = "delete_tab"
        const val DIALOG_ADD_TAB = "add_tab"
        const val DIALOG_EDIT_TAB = "edit_tab"
    }

    private lateinit var messagesViewModel: MessagesViewModel

    internal var allTabs = emptyList<Tab>()
    private lateinit var adapter: MessagesTabsAdapter

    private lateinit var pager: ViewPager2

    private fun toggleMenuItems(menu: Menu, except: MenuItem, vis: Boolean) {
        for (menuItem in menu.iterator()) {
            if (menuItem == except) {
                continue
            }

            menuItem.isVisible = vis
        }
    }

    override fun onCreateOptionsMenu(menu: Menu, inflater: MenuInflater) {
        // Inflate the menu; this adds items to the action bar if it is present.
        inflater.inflate(R.menu.messages_menu, menu)

        val searchItem = menu.findItem(R.id.app_bar_search)
        val search = searchItem.actionView as SearchView
        search.setOnSearchClickListener {
            this.toggleMenuItems(menu, searchItem, false)
        }
        search.setOnCloseListener {
            this.toggleMenuItems(menu, searchItem, true)

            false
        }
        search.setOnQueryTextListener(object : SearchView.OnQueryTextListener {
            override fun onQueryTextSubmit(query: String?): Boolean {
                return this.onQueryTextChange(query)
            }

            override fun onQueryTextChange(newText: String?): Boolean {
                val fragment = this@MessagesTabsFragment.childFragmentManager.findFragmentByTag("f${this@MessagesTabsFragment.pager.currentItem}")
                if (fragment !is MessagesFragment) {
                    return false
                }

                fragment.searchString(newText)
                return true
            }

        })
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            R.id.action_disconnect -> {
                val serversVm: ServersViewModel by this.getApp().viewModels()
                serversVm.connectionService.value?.disconnect()
            }
            R.id.add_tab -> {
                EditTextDialog.newInstance(Companion.DIALOG_ADD_TAB).show(this.childFragmentManager, "add_tab_name")
            }
            else -> {
                return false
            }
        }

        return true
    }

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        this.messagesViewModel = ViewModelProvider(this.requireActivity() as MainActivity).get(MessagesViewModel::class.java)
        val serversVm: ServersViewModel by this.getApp().viewModels()
        val binding = FragmentMessagesTabsBinding.inflate(inflater)

        val args: MessagesTabsFragmentArgs by this.navArgs()

        this.setHasOptionsMenu(true)

        serversVm.channel.observe(this.viewLifecycleOwner) {
            binding.messageInput.helperText = it ?: ""
        }

        binding.messageInput.apply {
            // don't auto-correct for command names
            this.editText?.addTextChangedListener { text ->
                if (text == null) {
                    return@addTextChangedListener
                }

                val editText = this.editText ?: return@addTextChangedListener

                val startsWithSlash = text.startsWith('/')

                val start = editText.selectionStart
                val end = editText.selectionEnd
                val firstSpace = text.indexOf(' ').let { if (it == -1) text.length else it }

                val inFirstWord = start <= firstSpace || end <= firstSpace

                val new = if (startsWithSlash && inFirstWord) {
                    // disable autocorrect if in the first word and it starts with a slash
                    editText.inputType and InputType.TYPE_TEXT_FLAG_AUTO_CORRECT.inv()
                } else {
                    // enable autocorrect otherwise
                    editText.inputType or InputType.TYPE_TEXT_FLAG_AUTO_CORRECT
                }

                // only change the input type if necessary (holding down backspace stops when input type changes)
                if (editText.inputType != new) {
                    editText.inputType = new
                }
            }

            fun sendMessage() {
                val view = this.editText ?: return
                val messageText = view.text.toString()
                this@MessagesTabsFragment.lifecycleScope.launch {
                    serversVm.connectionService.value?.messages?.send(ClientMessage(messageText))
                }
                view.setText("")
            }

            this.editText?.setOnEditorActionListener listener@{ _, actionId, _ ->
                if (actionId != EditorInfo.IME_ACTION_SEND) {
                    return@listener false
                }

                sendMessage()
                return@listener true
            }

            this.setEndIconOnClickListener {
                sendMessage()
            }

            this.setStartIconOnClickListener {
                AlertDialogFragment.newInstance(MessagesFragment.DIALOG_CHANNEL)
                    .show(this@MessagesTabsFragment.childFragmentManager, "channelChooser")
            }

            val editText = this.editText

            @SuppressLint("SetTextI18n")
            if (args.tellPlayer != null && args.tellWorld != null && editText != null) {
                val msg = "/tell ${args.tellPlayer}@${args.tellWorld} "
                editText.setText(msg)
                editText.setSelection(msg.length)
                editText.requestFocus()
                val imm = this.context.getSystemService(Context.INPUT_METHOD_SERVICE) as InputMethodManager
                // FIXME: showSoftInput is correct, but see the issue below
                //        https://stackoverflow.com/questions/10508363/show-keyboard-for-edittext-when-fragment-starts
                imm.toggleSoftInput(InputMethodManager.SHOW_FORCED, InputMethodManager.HIDE_NOT_ALWAYS)
//                imm.showSoftInput(editText, InputMethodManager.SHOW_IMPLICIT)
            }
        }

        // update the input whenever the connection changes
        serversVm.connected.observe(this.viewLifecycleOwner) {
            this.setUpInput(binding)
        }
        // update the input whenever the availability changes
        this.messagesViewModel.available.observe(this.viewLifecycleOwner) {
            this.setUpInput(binding)
        }
        // set up the input initially
        this.setUpInput(binding)

        this.adapter = MessagesTabsAdapter(this)
        binding.pager.adapter = this.adapter
        this.pager = binding.pager

        TabLayoutMediator(binding.tabLayout, binding.pager) { tab, position ->
            val currentTab = this.allTabs[position]

            tab.text = currentTab.name
            if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.M) {
                tab.view.isContextClickable = true
            }
            tab.view.setOnCreateContextMenuListener { menu, _, _ ->
                MenuInflater(this.context).inflate(R.menu.tab_menu, menu)
                menu.setHeaderTitle(tab.text)
                menu.findItem(R.id.delete).setOnMenuItemClickListener {
                    AlertDialogFragment.newInstance(Companion.DIALOG_DELETE_TAB, Bundle().apply {
                        this.putInt("tabId", currentTab.id)
                    }).show(this.childFragmentManager, "delete_tab")

                    true
                }
                menu.findItem(R.id.edit_name).setOnMenuItemClickListener {
                    EditTextDialog.newInstance(Companion.DIALOG_EDIT_TAB, Bundle().apply {
                        this.putString("currentName", currentTab.name)
                        this.putInt("id", currentTab.id)
                        this.putString("filters", currentTab.filters)
                    }).show(this.childFragmentManager, "edit_tab_name")

                    true
                }
                menu.findItem(R.id.edit_filters).setOnMenuItemClickListener {
                    FilterSelection.openExisting(this.childFragmentManager, currentTab.id)

                    true
                }
            }
        }.attach()

        this.messagesViewModel.allTabs.observe(this.viewLifecycleOwner) {
            this.allTabs = it
            this.adapter.notifyDataSetChanged()
        }

        return binding.root
    }

    private fun setUpInput(binding: FragmentMessagesTabsBinding) {
        val enabled: Boolean
        val hint: CharSequence
        val vm: ServersViewModel by this.getApp().viewModels()
        if (vm.connected.value == false) {
            enabled = false
            hint = this.getText(R.string.messages_not_connected)
        } else {
            if (this.messagesViewModel.available.value == true) {
                enabled = true
                hint = this.getText(R.string.messages_placeholder)
            } else {
                enabled = false
                hint = this.getText(R.string.messages_unavailable)
            }
        }

        binding.messageInput.apply {
            this.isEnabled = enabled
            this.isEndIconVisible = enabled
            this.editText?.hint = hint
        }
    }
}

class MessagesTabsAdapter(private val fragment: MessagesTabsFragment) : FragmentStateAdapter(fragment) {
    override fun getItemCount(): Int = this.fragment.allTabs.size

    override fun createFragment(position: Int): Fragment {
        if (position > this.fragment.allTabs.size) {
            throw Exception()
        }

        val tab = this.fragment.allTabs[position]

        return MessagesFragment().apply {
            this.arguments = Bundle().apply {
                this.putInt(MessagesFragment.ARG_TAB_ID, tab.id)
            }
        }
    }
}
