package io.annaclemens.xivchat.ui.servers

import android.os.Bundle
import android.text.InputFilter
import android.text.Spanned
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.AdapterView
import android.widget.PopupMenu
import androidx.activity.viewModels
import androidx.fragment.app.Fragment
import androidx.navigation.fragment.findNavController
import androidx.navigation.fragment.navArgs
import io.annaclemens.xivchat.R
import io.annaclemens.xivchat.databinding.FragmentNewServerBinding
import io.annaclemens.xivchat.model.SavedServer
import io.annaclemens.xivchat.util.AlertDialogFragment
import io.annaclemens.xivchat.util.getApp

class NewServerFragment : Fragment() {
    companion object {
        const val DIALOG_MISSING_FIELDS = "missing_fields"
    }

    private fun setServerType(binding: FragmentNewServerBinding, direct: Boolean) {
        val selection = if (direct) {
            0
        } else {
            1
        }
        binding.typeDropdown.setSelection(selection)

        val visibleIfDirect = if (direct) {
            View.VISIBLE
        } else {
            View.GONE
        }
        val goneIfDirect = if (direct) {
            View.GONE
        } else {
            View.VISIBLE
        }

        binding.directFields.visibility = visibleIfDirect
        binding.relayFields.visibility = goneIfDirect
    }

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View {
        val binding = FragmentNewServerBinding.inflate(inflater)

        val args: NewServerFragmentArgs by this.navArgs()

        val id = args.id
        this.getApp().supportActionBar?.setTitle(if (id <= 0) R.string.add_server_title else R.string.add_server_title_edit)

        val relay = args.relay

        val existingName = args.name
        if (existingName != null) {
            binding.name.setText(existingName)
        }

        this.setServerType(binding, !relay)

        if (relay) {
            val existingAuth = args.auth
            val existingTarget = args.target

            if (existingAuth != null) {
                binding.auth.setText(existingAuth)
            }
            if (existingTarget != null) {
                binding.target.editText?.setText(existingTarget)
            }
        } else {
            val existingHost = args.host
            val existingPort = args.port

            if (existingHost != null) {
                binding.host.setText(existingHost)
            }
            if (existingPort != 0) {
                binding.port.setText(existingPort.toString())
            }
        }

        if (id > 0) {
            binding.addButton.setText(R.string.add_server_button_edit)
        }

        binding.typeDropdown.onItemSelectedListener = object : AdapterView.OnItemSelectedListener {
            override fun onItemSelected(parent: AdapterView<*>?, view: View?, position: Int, id: Long) {
                val direct = position == 0
                this@NewServerFragment.setServerType(binding, direct)
            }

            override fun onNothingSelected(parent: AdapterView<*>?) {
                // no-op
            }
        }

        binding.port.apply {
            this.filters = arrayOf(object : InputFilter {
                override fun filter(source: CharSequence?, start: Int, end: Int, dest: Spanned?, dstart: Int, dend: Int): CharSequence? {
                    try {
                        // Remove the string out of destination that is to be replaced
                        var newVal = dest.toString().substring(0, dstart) + dest.toString().substring(dend, dest.toString().length)
                        newVal = newVal.substring(0, dstart) + source.toString() + newVal.substring(dstart, newVal.length)
                        val input = newVal.toInt()
                        if (this.isInRange(input)) {
                            return null
                        }
                    } catch (e: NumberFormatException) {
                    }
                    return ""
                }

                private fun isInRange(c: Int): Boolean = c in 1..65535
            })
        }

        binding.target.setEndIconOnClickListener {
            val popup = PopupMenu(this.context, binding.target)
            val trusted = this.getApp().preferences.all
                .filter { it.key.startsWith("trusted:") }
                .filter { it.value is Set<*> }
                .flatMap { it.value as Set<*> }
            for (key in HashSet(trusted)) {
                popup.menu.add(key.toString())
            }
            popup.setOnMenuItemClickListener {
                binding.target.editText?.setText(it.title)
                return@setOnMenuItemClickListener true
            }
            popup.show()
        }

        binding.addButton.setOnClickListener {
            val name = binding.name.text.toString()

            if (name.isBlank()) {
                AlertDialogFragment.newInstance(Companion.DIALOG_MISSING_FIELDS)
                    .show(this.childFragmentManager, "missing_fields")
                return@setOnClickListener
            }

            @Suppress("NAME_SHADOWING")
            val id = id.coerceAtLeast(0)
            val server: SavedServer

            if (binding.typeDropdown.selectedItemPosition == 0) {
                // direct
                val host = binding.host.text.toString()
                val portText = binding.port.text
                val port: Int = if (portText.isNullOrBlank()) {
                    14777
                } else {
                    portText.toString().toIntOrNull() ?: 14777
                }.coerceAtLeast(1).coerceAtMost(65535)

                if (host.isBlank()) {
                    AlertDialogFragment.newInstance(Companion.DIALOG_MISSING_FIELDS)
                        .show(this.childFragmentManager, "missing_fields")
                    return@setOnClickListener
                }

                server = SavedServer(
                    id = id,
                    name = name,
                    host = host,
                    port = port,
                    auth = null,
                    target = null,
                )
            } else {
                // relay
                val auth = binding.auth.text.toString().trim()
                val target = binding.target.editText!!.text.toString().trim()

                if (auth.isBlank() || target.isBlank()) {
                    AlertDialogFragment.newInstance(Companion.DIALOG_MISSING_FIELDS)
                        .show(this.childFragmentManager, "missing_fields")
                    return@setOnClickListener
                }

                server = SavedServer(
                    id,
                    name = name,
                    host = null,
                    port = null,
                    auth = auth,
                    target = target,
                )
            }

            val vm: ServersViewModel by this.getApp().viewModels()
            if (id <= 0) {
                vm.insert(server)
            } else {
                vm.update(server)
            }

            this.findNavController().navigateUp()
        }

        return binding.root
    }
}
