package io.annaclemens.xivchat.ui.servers.scan

import android.app.Service
import android.content.Context
import android.content.Intent
import android.net.wifi.WifiManager
import android.os.Binder
import android.os.Handler
import android.os.IBinder
import android.os.Looper
import android.os.Process
import androidx.activity.viewModels
import com.goterl.lazycode.lazysodium.interfaces.KeyExchange
import io.annaclemens.xivchat.MainActivity
import io.ktor.util.*
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.Job
import kotlinx.coroutines.ObsoleteCoroutinesApi
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.channels.produce
import kotlinx.coroutines.channels.sendBlocking
import kotlinx.coroutines.channels.ticker
import kotlinx.coroutines.launch
import kotlinx.coroutines.selects.select
import java.net.DatagramPacket
import java.net.InetAddress
import java.net.InetSocketAddress
import java.net.MulticastSocket
import java.nio.ByteBuffer

class ServerScannerService : Service() {
    inner class LocalBinder(private val _service: ServerScannerService) : Binder() {
        fun getService(): ServerScannerService {
            return this._service
        }
    }

    lateinit var app: MainActivity

    private val binder = this.LocalBinder(this)
    private lateinit var runner: ServerScanner
    private val handler = Handler(Looper.getMainLooper())

    override fun onBind(intent: Intent?): IBinder {
        return this.binder
    }

    fun spawn() {
        this.runner = ServerScanner(this)
        Thread(this.runner).start()
    }

    override fun onDestroy() {
        this.runner.cancel()
    }

    private class ServerScanner(private val service: ServerScannerService) : Runnable {
        private lateinit var job: Job
        private val cancelChannel = Channel<Unit>(1)

        fun cancel() {
            this.cancelChannel.sendBlocking(Unit)
            this.job.cancel()
        }

        @OptIn(KtorExperimentalAPI::class, ExperimentalUnsignedTypes::class, ExperimentalCoroutinesApi::class, ObsoleteCoroutinesApi::class)
        private suspend fun runAsync() {
            var run = true

            val ping = ticker(delayMillis = 5_000, initialDelayMillis = 0)

            val multicastAddr = InetAddress.getByName("224.0.0.147")
            val multicastPort = 17444

            val multicast = MulticastSocket(InetSocketAddress("0.0.0.0", multicastPort))
            multicast.joinGroup(multicastAddr)

            val incoming = CoroutineScope(Dispatchers.IO).produce {
                while (true) {
                    val packet = DatagramPacket(ByteArray(128), 128)
                    multicast.receive(packet)
                    this.send(packet)
                }
            }

            while (run) {
                select<Unit> {
                    this@ServerScanner.cancelChannel.onReceive {
                        run = false
                    }
                    ping.onReceive {
                        this@ServerScanner.service.handler.post {
                            val vm: ScannedServersViewModel by this@ServerScanner.service.app.viewModels()
                            vm.prune()
                        }

                        val payload = ByteArray(1)
                        payload[0] = 14
                        val packet = DatagramPacket(payload, payload.size)
                        packet.address = multicastAddr
                        packet.port = multicastPort
                        multicast.send(packet)
                    }
                    incoming.onReceive {
                        val buffer = ByteBuffer.wrap(it.data).apply {
                            this.limit(it.length)
                        }

                        if (buffer.get() != 14.toByte() || !buffer.hasRemaining()) {
                            return@onReceive
                        }

                        val strLen = buffer.get().toInt()
                        val strBytes = ByteArray(strLen)
                        buffer.get(strBytes)
                        val name = strBytes.decodeToString()

                        val port = buffer.short.toUShort()

                        val key = ByteArray(KeyExchange.PUBLICKEYBYTES)
                        buffer.get(key)

                        val server = ScannedServer(
                            playerName = name,
                            address = it.address.hostAddress,
                            port = port,
                            publicKey = key,
                        )

                        this@ServerScanner.service.handler.post {
                            val vm: ScannedServersViewModel by this@ServerScanner.service.app.viewModels()
                            vm.addServer(server)
                        }
                    }
                }
            }

            multicast.close()
        }

        override fun run() {
            Process.setThreadPriority(Process.THREAD_PRIORITY_BACKGROUND)

            this.job = CoroutineScope(Dispatchers.Default).launch {
                val wifi = this@ServerScanner.service.applicationContext.getSystemService(Context.WIFI_SERVICE) as WifiManager?
                val multicastLock = wifi?.createMulticastLock("xivchat-scan-multicast")
                multicastLock?.acquire()

                try {
                    this@ServerScanner.runAsync()
                } catch (ex: Exception) {
                    ex.printStackTrace()
                }

                multicastLock?.release()
            }
        }
    }
}
